#include "GridP.h"

/* Error Messages */
#define MESSAGE1 "There must be at least one row."
#define MESSAGE2 "There must be at least one column."
#define MESSAGE3 "The row position can not be greater that the number of rows."
#define MESSAGE4 "The column position can not be greater that the number of columns."
#define MESSAGE5 "The row and column position is already taken.\n  Defaulting to next available position."
#define MESSAGE6 "The grid is full.  The number of rows or columns must be expaneded."
#define MESSAGE7 "The row and column position is already taken.\n  Retaining old position."

/********    Static Function Declarations    ********/

/* core class procedures */
static void ClassInitialize( 
                        void );
static void ClassPartInitialize( 
                        WidgetClass );
static void Initialize( 
                        Widget request,
                        Widget new_w,
                        ArgList args,
                        Cardinal *num_args );
static void Destroy( 
                        Widget w );
static void Resize( 
                        Widget w );
static void Redisplay( 
                        Widget w,
                        XEvent *event,
                        Region region );
static Boolean SetValues( 
                        Widget old_w,
                        Widget request,
                        Widget new_w,
                        ArgList args,
                        Cardinal *num_args );
static XtGeometryResult QueryGeometry(
                        Widget wid,
                        XtWidgetGeometry *request,
                        XtWidgetGeometry *ret );

/* composite class procedures */
static XtGeometryResult GeometryManager(
                        Widget w,
                        XtWidgetGeometry *request,
                        XtWidgetGeometry *reply );
static void ChangeManaged(
                        Widget wid );

/* constraint procedures */
static void ConstraintInitialize(
                        Widget rw,
                        Widget nw,
                        ArgList args,
                        Cardinal *num_args );
static Boolean ConstraintSetValues(
                        Widget cw,
                        Widget rw,
                        Widget nw,
                        ArgList args,
                        Cardinal *num_args );

/* grid class procedures */
static void UpdateGridDimensions(
                        Widget w,
                        Dimension * width,
                        Dimension * height );
static void AddChildToGrid(
                        Widget parent,
                        Widget child,
                        Dimension row_pos,
                        Dimension col_pos );
static void LayoutChildren(
                        Widget w );
static void DrawGridLines(
                        Widget w );
static void EraseGridLines(
                        Widget w );

/* internal procedures */
static Dimension GetColumnWidth(
                        XsmGridWidget gw,
                        int col_pos );
static Dimension GetRowHeight(
                        XsmGridWidget gw,
                        int row_pos );
static Boolean SetChildHorizPosition(
                        XsmGridWidget parent,
                        Widget child, 
                        XsmGridConstraintPart * grid_info,
                        Dimension row,
                        Dimension start_col );
static Boolean SetChildVertPosition(
                        XsmGridWidget parent,
                        Widget child, 
                        XsmGridConstraintPart * grid_info,
                        Dimension start_row,
                        Dimension col );
static Boolean SetChildPosition(
                        XsmGridWidget parent,
                        Widget child, 
                        XsmGridConstraintPart * grid_info,
                        Dimension start_row,
                        Dimension start_col );
static void SetInvGC(
                        XsmGridWidget gw,
                        GC gc );
static void SetNormGC(
                        XsmGridWidget gw,
                        GC gc );
static void LoadGC( 
                        XsmGridWidget gw,
                        Pixel fg,
                        Pixel bg );
static void ModifyGridLines(
                        XsmGridWidget gw,
#if NeedWidePrototypes
                        int draw_grid );
#else
                        Boolean draw_grid );
#endif /* NeedWidePrototypes */


/* Macro to get constraint resource values */
#define GridInfo(w)     ((XsmGridConstraintPtr)(w)->core.constraints)

/* Macro for stack allocation */
#define StackAlloc(size, stack_array) \
                   ((size) <= sizeof(stack_array) ? (XtPointer)(stack_array) \
                                                      : XtMalloc((unsigned)(size)))
#define StackFree(pointer, stack_array) \
    if ((pointer) != ((XtPointer)(stack_array))) XtFree((char *)pointer);

#define MAXDIMENSION        ((1 << 16)-1)


/********    End Static Function Declarations    ********/


static char *HorizontalAlignmentNames[] = {"align_center","align_left",
                                           "align_right"};
static char *VerticalAlignmentNames[] = {"align_center","align_top",
                                         "align_bottom"};

#define NUM_NAMES( list )        (sizeof( list) / sizeof( char *))


/* This is the static initialization of the resources for the
 * Grid widget.  Note: the synthetic resources are also included
 * this resource list as well as in the synthetic resource list.
 * The synthetic resource list includes additional information 
 * about the synthetis resource.
 */

static XtResource resources[] = {

        {
          XsmNcolumns,
          XsmCColumns,
          XmRDimension,
          sizeof(Dimension),
          XtOffset (XsmGridWidget, grid.columns),
          XmRImmediate, (XtPointer) 1
        },

        {
          XsmNrows,
          XsmCRows,
          XmRDimension,
          sizeof(Dimension),
          XtOffset (XsmGridWidget, grid.rows),
          XmRImmediate, (XtPointer) 1
        },

        {
          XsmNmarginWidth,
          XsmCMarginWidth,
          XmRHorizontalDimension,
          sizeof(Dimension),
          XtOffset (XsmGridWidget, grid.margin_width),
          XmRImmediate, (XtPointer) 5
        },

        {
          XsmNmarginHeight,
          XsmCMarginHeight,
          XmRVerticalDimension,
          sizeof(Dimension),
          XtOffset (XsmGridWidget, grid.margin_height),
          XmRImmediate, (XtPointer) 5
        },

        {
          XsmNhorizontalSpacing,
          XsmCHorizontalSpacing,
          XmRHorizontalDimension,
          sizeof(Dimension),
          XtOffset (XsmGridWidget, grid.h_spacing),
          XmRImmediate, (XtPointer) 5
        },

        {
          XsmNverticalSpacing,
          XsmCVerticalSpacing,
          XmRVerticalDimension,
          sizeof(Dimension),
          XtOffset (XsmGridWidget, grid.v_spacing),
          XmRImmediate, (XtPointer) 5
        },

        {
          XsmNhorizontalAlignment,
          XsmCHorizontalAlignment,
          XsmRHorizontalAlignment,
          sizeof(unsigned char),
          XtOffset (XsmGridWidget, grid.horiz_alignment),
          XmRImmediate, (XtPointer) XsmALIGN_CENTER
        },

        {
          XsmNverticalAlignment,
          XsmCVerticalAlignment,
          XsmRVerticalAlignment,
          sizeof(unsigned char),
          XtOffset (XsmGridWidget, grid.vert_alignment),
          XmRImmediate, (XtPointer) XsmALIGN_MIDDLE
        },

        {
          XsmNshowGridLines,
          XsmCShowGridLines,
          XmRBoolean,
          sizeof(Boolean),
          XtOffset (XsmGridWidget, grid.show_grid_lines),
          XmRImmediate, (XtPointer) False
        },
};
          
/* Synthetic resources change their values dynamically.
 * As in Motif, our dimension resources change dynamically
 * based on the XmNunitType resource.  Foe example, if the unit type
 * is in 1000ths of an inch, the conversion routines will modify
 * the resource values to convert the resource value in the given
 * unit type into a number of pixels for use in displaying the
 * the widget in the stated dimensions.
 */
static XmSyntheticResource syn_resources[] =
{
   {
      XmNmarginWidth,
      sizeof (Dimension),
      XtOffset( XsmGridWidget, grid.margin_width),
      _XmFromHorizontalPixels,
      _XmToHorizontalPixels
   },

   {
      XmNmarginHeight,
      sizeof (Dimension),
      XtOffset( XsmGridWidget, grid.margin_height),
      _XmFromVerticalPixels,
      _XmToVerticalPixels,
   },

   {
      XmNhorizontalSpacing,
      sizeof (Dimension),
      XtOffset( XsmGridWidget, grid.h_spacing),
      _XmFromHorizontalPixels,
      _XmToHorizontalPixels
   },

   { 
      XmNverticalSpacing,
      sizeof (Dimension),
      XtOffset( XsmGridWidget, grid.v_spacing),
      _XmFromVerticalPixels,
      _XmToVerticalPixels,
   },
};

/* Constraint resources for children of the Grid widget */
static XtResource constraint_resources[] =
{

    {
       XsmNrowPosition,
       XsmCRowPosition,
       XmRPosition,
       sizeof(Position),
       XtOffsetOf( struct _XsmGridConstraintRec, grid.row_pos),
       XmRImmediate, (XtPointer) MAXDIMENSION
    },

    {
       XsmNcolumnPosition,
       XsmCColumnPosition,
       XmRPosition,
       sizeof(Position),
       XtOffsetOf( struct _XsmGridConstraintRec, grid.col_pos),
       XmRImmediate, (XtPointer) MAXDIMENSION
    },
};


/* This is the initialization of the class record.  The core class part
 * is initialized first, followed by the Composite, Constraint, Manager,
 * and Grid class parts.
 */
externaldef (xsmgridclassrec) XsmGridClassRec xsmGridClassRec = {
   {            /* core_class fields       */
      (WidgetClass) &xmManagerClassRec,         /* superclass            */
      "XsmGrid",                                /* class_name            */
      sizeof(XsmGridRec),                       /* widget_size           */
      ClassInitialize,                          /* class_initialize      */
      ClassPartInitialize,                      /* class_part_initialize */
      FALSE,                                    /* class_inited          */
      Initialize,                               /* initialize            */
      NULL,                                     /* initialize_hook       */
      XtInheritRealize,                         /* realize               */
      NULL,                                     /* actions               */
      0,                                        /* num_actions           */
      resources,                                /* resources             */
      XtNumber(resources),                      /* num_resources         */
      NULLQUARK,                                /* xrm_class             */
      TRUE,                                     /* compress_motion       */
      XtExposeCompressMaximal,                  /* compress_exposure     */
      TRUE,                                     /* compress_enterleave   */
      FALSE,                                    /* visible_interest      */
      Destroy,                                  /* destroy               */
      Resize,                                   /* resize                */
      Redisplay,                                /* expose                */
      SetValues,                                /* set_values            */
      NULL,                                     /* set_values_hook       */
      XtInheritSetValuesAlmost,                 /* set_values_almost     */
      NULL,                                     /* get_values_hook       */
      NULL,                                     /* accept_focus          */
      XtVersion,                                /* version               */
      NULL,                                     /* callback_private      */
      NULL,                                     /* tm_table              */
      QueryGeometry,                            /* query_geometry        */
      XtInheritDisplayAccelerator,              /* display accel         */
      NULL,                                     /* extension             */
   },

   {            /* composite_class fields  */
      GeometryManager,                          /* geometry_manager      */
      ChangeManaged,                            /* change_managed        */
      XtInheritInsertChild,                     /* insert_child          */
      XtInheritDeleteChild,                     /* delete_child          */
      NULL,                                     /* extension             */
   },

   {            /* constraint_class fields */
      constraint_resources,                     /* resource list         */
      XtNumber(constraint_resources),           /* num resources         */
      sizeof (XsmGridConstraintRec),            /* constraint size       */
      ConstraintInitialize,                     /* init proc             */
      NULL,                                     /* destroy proc          */
      ConstraintSetValues,                      /* set values proc       */
      NULL,                                     /* extension             */
   },

   {            /* manager_class fields    */
      XtInheritTranslations,                    /* translations          */
      syn_resources,                            /* syn resources         */
      XtNumber(syn_resources),                  /* num syn_resources     */
      NULL,                                     /* syn_cont_resources    */
      0,                                        /* num_syn_cont_resources*/
      XmInheritParentProcess,                   /* parent_process        */
      NULL,                                     /* extension             */
   },

   {           /* grid_class fields        */
      UpdateGridDimensions,			/* update_grid           */
      AddChildToGrid,                           /* add_child             */
      LayoutChildren,                           /* layout_children       */
      DrawGridLines,			        /* draw_grid             */
      EraseGridLines,			        /* erase_grid            */
      NULL,                                     /* extension             */
   }
};

/* enternal definition of the grid class record */
externaldef(xsmgridwidgetclass) WidgetClass xsmGridWidgetClass =
                                         (WidgetClass) &xsmGridClassRec;

/****************************************************************
 * Core Class Methods
 ****************************************************************/

/************************************************************************
 *
 *  ClassInitialize() -
 *      The widget class initialization routine.  
 *      Representation types are registered here for new types
 *      not previously registered by Motif or the Intrinsics.
 *
 ************************************************************************/
static void
ClassInitialize( void )
{
   /* Register the representation types.  This can only be used with
    * resources that are enum types.  Other types must use
    * XtSetTypeConverter().
    */
    XmRepTypeRegister( XsmRHorizontalAlignment, HorizontalAlignmentNames, NULL,
                                           NUM_NAMES(HorizontalAlignmentNames));
    XmRepTypeRegister( XsmRVerticalAlignment, VerticalAlignmentNames, NULL,
                                           NUM_NAMES(VerticalAlignmentNames));
}

^L
/************************************************************************
 *
 *  ClassPartInitialize() -
 *       The initialization routine for class parts.  It is called for the
 *       initialization of this class and each time a new subclass is created
 *       for this widget class.  It is called only once per subclass.
 *
 ************************************************************************/
 
FooBar( void )
{
    XsmGridWidgetClass gc = (XsmGridWidgetClass) wc;
    XsmGridWidgetClass sc = (XsmGridWidgetClass) wc->core_class.superclass;

   /* assign procedures to the classes pointers that inherit these procedures */

    if (gc->grid_class.add_child == XsmInheritAddChild)
       gc->grid_class.add_child = sc->grid_class.add_child;

    if (gc->grid_class.layout_children == XsmInheritLayoutChildren)
       gc->grid_class.layout_children = sc->grid_class.layout_children;

    if (gc->grid_class.draw_grid == XsmInheritDrawGrid)
       gc->grid_class.draw_grid = sc->grid_class.draw_grid;

    if (gc->grid_class.erase_grid == XsmInheritEraseGrid)
       gc->grid_class.erase_grid = sc->grid_class.erase_grid;
}

/************************************************************************
 *
 *  Initialize() -
 *      The main widget instance initialization routine.  Initialize the
 *        instance variables here, and verify the resources values.
 *
 ************************************************************************/

/* ARGSUSED */
static void
Initialize( Widget request, Widget new_w, ArgList args, Cardinal *num_args )
{
   XsmGridWidget new_m = (XsmGridWidget) new_w;
   XsmGridWidget req_m = (XsmGridWidget) request;
   int i;

  /* verify resource values */
   if (new_m->grid.rows < 1)
   {
      _XmWarning( (Widget) new_w, MESSAGE1);
      new_m->grid.rows = 1;
   }

   if (new_m->grid.columns < 1)
   {
      _XmWarning( (Widget) new_w, MESSAGE2);
      new_m->grid.columns = 1;
   }

  /* Ensure a minimum size */

   if (req_m->core.width == 0)
     new_m->core.width = (2 * new_m->grid.margin_width) +
                         (new_m->grid.h_spacing ?
                          new_m->grid.h_spacing : 10) *
                          (new_m->grid.columns - 1);
                         
   if (new_m->core.height == 0)
     new_m->core.height = (2 * new_m->grid.margin_height) +
                          (new_m->grid.v_spacing ?
                           new_m->grid.v_spacing : 10) *
                           (new_m->grid.rows - 1);

  /* Allocate space for the list of row heights */
   new_m->grid.row_height = (Dimension *) XtCalloc(new_m->grid.rows,
                                                    sizeof(Dimension));

  /* Allocate space for the list of column widgets */
   new_m->grid.col_width = (Dimension *) XtCalloc(new_m->grid.columns,
                                                    sizeof(Dimension));

  /* Allocate space for the grid of widgets */
   new_m->grid.widget_grid = (WidgetList *) XtMalloc(sizeof(WidgetList) * 
                                                   new_m->grid.rows);

   for (i = 0; i < new_m->grid.rows; i++)
   {
       new_m->grid.widget_grid[i] = (WidgetList)XtCalloc(new_m->grid.columns,
                                                          sizeof(Widget));
   }

  /* Save original margins for use in Resize() */
   new_m->grid.orig_margin_width = new_m->grid.margin_width;
   new_m->grid.orig_margin_height = new_m->grid.margin_height;

   new_m->grid.lines_visible = False;

   LoadGC(new_m, new_m->core.background_pixel, new_m->manager.foreground);
}



/************************************************************************
 *
 *  Destroy() -
 *      Clean up allocated resources when the widget is destroyed.
 *        Anything that is malloc'd byt the widget should be freed here.
 *
 ************************************************************************/

static void
Destroy( Widget w )
{
   XsmGridWidget gw = (XsmGridWidget) w;
   int i;
   
  /* Free row_height and column_width arrays */
   XtFree((char *)gw->grid.row_height);
   XtFree((char *)gw->grid.col_width);

  /* Free two dimensional widget_grid array */
   for (i = 0; i < gw->grid.rows; i++)
      XtFree((char *)gw->grid.widget_grid[i]);
   XtFree((char *)gw->grid.widget_grid);
}


/************************************************************************
 *
 *  Resize() -
 *      Redraw according to size changes.  Adjust only the margins to
 *  account for growth.  Clip children if the dimensions are too small.
 *
 ************************************************************************/

static void
Resize( Widget w )
{
   XsmGridWidget gw = (XsmGridWidget) w;
   XsmGridWidgetClass gc = (XsmGridWidgetClass) XtClass(w);
   Dimension width, height, save_margin_width, save_margin_height;
   
  /* get grid height and width */
   (*gc->grid_class.update_grid)(w, &width, &height);

  /* add in the margin widths and inter-column spacing */
   width += ((2 * gw->grid.orig_margin_width) +
                      (gw->grid.h_spacing * (gw->grid.rows - 1)));

  /* add in the margin widths and inter-column spacing */
   height += ((2 * gw->grid.orig_margin_height) +
                       (gw->grid.v_spacing * (gw->grid.columns - 1)));

  /* save current margins */
   save_margin_width = gw->grid.margin_width;
   save_margin_height = gw->grid.margin_height;
   
  /* set to resource settings for margins */
   gw->grid.margin_width = gw->grid.orig_margin_width;
   gw->grid.margin_height = gw->grid.orig_margin_height;

  /* if there is a resize growth, reset the margins */
   if (gw->core.width > width)
      gw->grid.margin_width += (gw->core.width - width)/2;

   if (gw->core.height > height)
      gw->grid.margin_height += (gw->core.height - height)/2;

  /* relayout if the margins changed */
   if (gw->grid.margin_width != save_margin_width ||
       gw->grid.margin_height != save_margin_height)
      (*gc->grid_class.layout_children)(w);
}


/************************************************************************
 *
 *  Redisplay() -
 *      Redraw region that gets exposed.
 *
 ************************************************************************/

static void
Redisplay( Widget w, XEvent *event, Region region )
{
   XsmGridWidget gw = (XsmGridWidget) w;
   XsmGridWidgetClass gc = (XsmGridWidgetClass) XtClass(w);

   /*  Redisplay gadgets.  */
    _XmRedisplayGadgets(w, event, region);

    if (gw->grid.show_grid_lines)
       (*gc->grid_class.draw_grid)(w);
    else if (gw->grid.lines_visible)
       (*gc->grid_class.erase_grid)(w);
}



/************************************************************************
 *
 *  SetValues() -
 *      The procedure is used to verify changes to the resource values.
 *      If the resource changes effect the visuals of the widget, the
 *      return value of the procedure should be set to True indicating
 *      that the expose procedure will be called to redraw the widget.
 *      Any changes to the widget's geometry in this procedure will
 *      generate a geometry request to it's parent.
 *
 ************************************************************************/

/* ARGSUSED */
static Boolean
SetValues( Widget old_w, Widget request, Widget new_w, ArgList args,
           Cardinal *num_args )
{
   XsmGridWidget new_m = (XsmGridWidget) new_w;
   XsmGridWidget old_m = (XsmGridWidget) old_w;
   XsmGridWidgetClass gc = (XsmGridWidgetClass) XtClass(new_w);
   XsmGridConstraintPart * grid_info;
   Boolean redisplay = False;
   Dimension col_pos;
   Dimension row_pos;
   int i, j;

  /* verify resource values */
   if (new_m->grid.rows < 1)
   {
      _XmWarning( (Widget) new_w, MESSAGE1);
      new_m->grid.rows = old_m->grid.rows;
   }

   if (new_m->grid.columns < 1)
   {
      _XmWarning( (Widget) new_w, MESSAGE2);
      new_m->grid.columns = old_m->grid.columns;
   }

  /* Reset arrays if rows or columns change. */
   if (new_m->grid.rows != old_m->grid.rows)
     /* Reallocate space for the list of row heights */
      new_m->grid.row_height = (Dimension *)XtRealloc((XtPointer)
                                                        new_m->grid.row_height,
                                                        sizeof(Dimension) *
                                                        new_m->grid.rows);

   if (new_m->grid.columns != old_m->grid.columns)
     /* Reallocate space for the list of column widgets */
      new_m->grid.col_width = (Dimension *) XtRealloc((XtPointer)
                                                        new_m->grid.col_width,
                                                        sizeof(Dimension) *
                                                        new_m->grid.columns);
   if (new_m->grid.rows != old_m->grid.rows ||
       new_m->grid.columns != old_m->grid.columns)
   {
     /* Free up old grid of widgets */
      for (i = 0; i < old_m->grid.rows; i++)
      {
         XtFree((char *)new_m->grid.widget_grid[i]);
      }

     /* Reallocate space for the grid of widgets */
      new_m->grid.widget_grid = (WidgetList *)XtRealloc((XtPointer)
                                                        new_m->grid.widget_grid,
                                                        sizeof(WidgetList) * 
                                                        new_m->grid.rows);

     /* Reallocate space for the grid of widgets */
      for (i = 0; i < new_m->grid.rows; i++)
         new_m->grid.widget_grid[i] = (WidgetList) XtCalloc(new_m->grid.columns,
                                                            sizeof(Widget));
      redisplay = True;
   }

  /* Reset children's positions in the position arrays when the
   * rows or columns change.
   */
   if (new_m->grid.rows < old_m->grid.rows ||
       new_m->grid.columns < old_m->grid.columns)
   {
      Widget child;
      Dimension width, height;

      for (i = 0; i < new_m->composite.num_children; i++)
      {
            child = new_m->composite.children[i];
            grid_info = &(GridInfo(child)->grid);
            
            row_pos = grid_info->orig_row_pos;
            col_pos = grid_info->orig_col_pos;

            (*gc->grid_class.add_child)(new_w, child, row_pos, col_pos);
      }

     /* get grid height and width */
      (*gc->grid_class.update_grid)(new_w, &width, &height);

     /* Update width and height.  If they change this will automatically
      * cause a geometry request to the grids parent.
      */
      new_m->core.width = width + ((2 * new_m->grid.margin_width) +
                          (new_m->grid.h_spacing * (new_m->grid.rows - 1)));

      new_m->core.height = height + ((2 * new_m->grid.margin_height) +
                           (new_m->grid.v_spacing * (new_m->grid.columns - 1)));

      (*gc->grid_class.layout_children)(new_w);
   }

  /* Save changes in margins for use in Resize() */
   new_m->grid.orig_margin_width = new_m->grid.margin_width;
   new_m->grid.orig_margin_height = new_m->grid.margin_height;

   if (new_m->core.background_pixel != old_m->core.background_pixel ||
       new_m->manager.foreground != old_m->manager.foreground)
   {
      LoadGC(new_m, new_m->core.background_pixel, new_m->manager.foreground);
      redisplay = True;
   }

   if (new_m->grid.show_grid_lines != old_m->grid.show_grid_lines)
      redisplay = True;

   return (redisplay);
}


/************************************************************************
 *
 *  QueryGeometry() -
 *      When manager widget have a query geometry procedure it should
 *      make query geometry requests to their children to calculate its
 *      preferred geometry.  Once this is done, set the reply fields
 *        and determine the appropriate XtGeometryResult to return.
 *
 ************************************************************************/

static XtGeometryResult
QueryGeometry( Widget wid, XtWidgetGeometry *request, XtWidgetGeometry *reply )
{
   XsmGridWidget gw = (XsmGridWidget) wid;
   Widget child;
   Dimension col_stack[50];
   Dimension *max_col_width;
   Dimension row_height, width = 0, height = 0;
   XtWidgetGeometry desired;
   int i, j;

  /* Set up an array to save the max width of each column.
   * The StackAlloc macro will allocate space if the space
   * required is more that was allocated on the stack. 
   */ 
   max_col_width = (Dimension *) StackAlloc(gw->grid.columns, col_stack);

  /* initialize the max_col_width array to zero */
   for (j = 0; j < gw->grid.columns; j++)
       max_col_width[j] = 0;

  /* Go through the rows caluclating the height of each row
   * from the desired height of each child.  Also use the
   * values from the QueryGeometry to figure out the max width
   * of each column.
   */
   for (i = 0; i < gw->grid.rows; i++)
   {
       row_height = 0;

       for (j = 0; j < gw->grid.columns; j++)
       {
           child = gw->grid.widget_grid[i][j];

           if (child != NULL)
           {
              XtQueryGeometry(child, NULL, &desired);

              if (row_height < desired.height)
                 row_height = desired.height;
           
              if (max_col_width[j] < desired.width)
                 max_col_width[j] = desired.width;
           }
       }

       height += row_height;
   }

   for (j = 0; j < gw->grid.columns; j++)
       width += max_col_width[j];

  /* Free the array.  The StackFree macro only free space if it
   * was not allocated on the stack.
   */
   StackFree(max_col_width, col_stack);

  /* Set up the reply to the query */
   reply->width = width + ((2 * gw->grid.margin_width) +
                           (gw->grid.h_spacing * (gw->grid.rows - 1)));

   reply->height = height + ((2 * gw->grid.margin_height) +
                             (gw->grid.v_spacing * (gw->grid.columns - 1)));

   reply->request_mode = CWWidth | CWHeight;
   
  /* XtGeometryYes indicates that the requested geometry is the same
   * as reply geometry.
   */
   if ((request->request_mode & CWWidth && request->width == reply->width) &&
       (request->request_mode & CWHeight && request->height == reply->height))
      return XtGeometryYes;

  /* XtGeometryNo indicates that our current geometry is the same
   * as reply geometry.
   */
   if (reply->width == gw->core.width &&
       reply->height == gw->core.height)
      return XtGeometryNo;

  /* XtGeometryAlmost indicates that at least one of the fields in the 
   * reply geometry differs from the request geometry.
   */
   return XtGeometryAlmost;
}

/****************************************************************
 * Composite Class Methods
 ****************************************************************/


/************************************************************************
 *
 *  GeometryManager() -
 *      Sets up geometry negotiations between the grid widget, the child
 *        requesting a change in geometry, and the grid widget's parent.
 *        This function determines whether a child is allow to grow or
 *        shrink. (Other Geometry Managers also allow their children to
 *        reposition themselves and change their border widths.)
 *
 ************************************************************************/

static XtGeometryResult
GeometryManager( Widget w, XtWidgetGeometry *request, XtWidgetGeometry *reply )
{
   XsmGridWidget gw = (XsmGridWidget) XtParent(w);
   XsmGridConstraintPart * grid_info = &(GridInfo(w)->grid);
   XsmGridWidgetClass gc = (XsmGridWidgetClass) XtClass((Widget)gw);
   Dimension width = 0, height = 0;
   Dimension save_width, save_height;
   XtGeometryResult result = XtGeometryYes;
   XtWidgetGeometry geo_request;
   XtWidgetGeometry geo_reply;
   Boolean set_almost = False;
   Boolean make_request = False;
   XtGeometryMask mask = request->request_mode;
   int i, j;

  /* reject attempts that do not want to adjust height or width */
   if (!(mask & (CWWidth | CWHeight)))
        return XtGeometryNo;

  /* Adjustments to x, y or border width are not allowed. 
   * Continue with the request for height and width but be
   * sure to return XtGeometryAlmost if x, y or border_with
   * are requested.
   */
   if (mask & CWX && request->x != w->core.x)
   {
      reply->x = w->core.x;
      reply->request_mode |= CWX;
      set_almost = True;
   }

   if (mask & CWY && request->y != w->core.y)
   {
      reply->y = w->core.y;
      reply->request_mode |= CWY;
      set_almost = True;
   }

   if (mask & CWBorderWidth && request->border_width != w->core.border_width)
   {
      reply->y = w->core.border_width;
      reply->request_mode |= CWBorderWidth;
      set_almost = True;
   }

  /* Check for queryonly or requests other than width and height.
   * If the request more than just width and height, we must
   * return GeometryAlmost. This means the request to our parent must be
   * a queryonly request to prevent our parent from reconfiguring our geometry.
   */
   if ((mask & XtCWQueryOnly) || set_almost)
      geo_request.request_mode = XtCWQueryOnly;
   else
      geo_request.request_mode = 0 ;

  
   if (mask & CWWidth)
   {
      /* If the request width causes the column to grow or
       * the request width may cause the column to shrink
       * see if the change will affect the grid's geometry.
       * Otherwise except the request and change the childs width.
       */
      if (request->width > gw->grid.col_width[grid_info->col_pos] ||
          (w->core.width == gw->grid.col_width[grid_info->col_pos] &&
           request->width < gw->grid.col_width[grid_info->col_pos]))
      {
        /* Save the child's width in case of a failure to acquire a new
         * width from the grid's parent or in case the request is a query.
         * Set the width to the requested value  for use in computations.
         */
         save_width = w->core.width;
         w->core.width = request->width;

        /* find total width for the columns */
         for (j = 0; j < gw->grid.columns; j++)
         {
            width += GetColumnWidth(gw, j);
         }

        /* add in the margin widths and inter-column spacing */
         width += ((2 * gw->grid.margin_width) +
                         (gw->grid.h_spacing * (gw->grid.rows - 1)));

        /* if the computed width is not equal to the current grid width
         * make the request to the parent of the grid widget for a new size
         */
         if (gw->core.width != width)
         {
            make_request = True;
            geo_request.request_mode |= CWWidth;
            geo_request.width = width;
         }
      } else
         w->core.width = request->width;
   }

   if (mask & CWHeight)
   {
      /* If the request height causes the row to grow or
       * the request height may cause the row to shrink
       * see if the change will affect the grid's geometry.
       * Otherwise except the request and change the child's height.
       */
      if (request->height > gw->grid.row_height[grid_info->row_pos] ||
          (w->core.height == gw->grid.row_height[grid_info->row_pos] &&
           request->height < gw->grid.row_height[grid_info->row_pos]))
      {
        /* Save the child's width in case of a failure to acquire a new
         * width from the grid's parent or in case the request is a query.
         * Set the width to the requested value  for use in computations.
         */
         save_height = w->core.height;
         w->core.height = request->height;

        /* find total width for the columns */
         for (i = 0; i < gw->grid.rows; i++)
         {
            height += GetRowHeight(gw, i);
         }

        /* add in the margin widths and inter-column spacing */
         height != ((2 * gw->grid.margin_height) +
                          (gw->grid.v_spacing * (gw->grid.columns - 1)));

        /* if the computed width is not equal to the current grid width
         * make the request to the parent of the grid widget for a new size
         */
         if (gw->core.height != height)
         {
            make_request = True;
            geo_request.request_mode |= CWHeight;
            geo_request.height = height;
         }
      } else
         w->core.height = request->height;
   }

   if (make_request)
   {
     /* Ask the grid's parent if it is allowed to change size */
      result = XtMakeGeometryRequest((Widget) gw, &geo_request, &geo_reply);

     /* request failed */
      if (result == XtGeometryNo)
      {
         w->core.width = save_width;
         w->core.height = save_height;
         return(result);
      }
      
     /* part of the request failed  and part succeeded */
      if (result == XtGeometryAlmost)
      {
         reply->request_mode = mask;

        /* if the child is making a request to change its width */
         if (mask & CWWidth)
         {
           /* if the grid also needed to make a request to change its width */
            if (geo_request.request_mode & CWWidth)
            {
              /* if the request will be granted for width */
               if (geo_request.width == reply->width)
                  reply->width = w->core.width;
               else
                 /* if the request failed reset the width to the difference
                  * between the parent's reply and request.
                  */
                  reply->width = w->core.width +
                                (reply->width - geo_request.width);
            }
            else
            { 
              /* if the parent didn't need to make a request for a change
               * in width, the new child width must be o.k.
               */
               reply->width = w->core.width;
            }

            w->core.width = save_width;
         }
               
        /* if the child is making a request to change its height */
         if (mask & CWHeight)
         {
           /* if the grid also needed to make a request to change its height */
            if (geo_request.request_mode & CWHeight)
            {
              /* if the request will be granted for height */
               if (geo_request.height == reply->height)
                  reply->height = w->core.height;
               else
                 /* if the request failed reset the height to the difference
                  * between the parent's reply and request.
                  */
                  reply->height = w->core.height +
                                (reply->height - geo_request.height);
            }
            else
            { 
              /* if the parent didn't need to make a request for a change
               * in height, the new child height must be o.k.
               */
               reply->height = w->core.height;
            }

            w->core.height = save_height;
         }
      }
      return (XtGeometryAlmost);
   }

  /* if there was a change in x, y, of border_width that we couldn't grant */
   if (set_almost)
      result = XtGeometryAlmost;

   if ((mask & XtCWQueryOnly))
   {
      if (mask & CWWidth)
         w->core.width = save_width;
      if (mask & CWHeight)
         w->core.height = save_height;
   }
   else
   {
     /* If we succeeded, change the row_height and col_width arrays
      * and relayout the children.
      */
      if (result == XtGeometryYes)
      {
         gw->grid.row_height[grid_info->row_pos] =
                                   GetRowHeight(gw, grid_info->row_pos);
         gw->grid.col_width[grid_info->col_pos] =
                                   GetColumnWidth(gw, grid_info->col_pos);
         (*gc->grid_class.layout_children)((Widget) gw);
      }
   }
      
   
   return (result);
}


/************************************************************************
 *
 *  ChangeManaged
 *      Something changed in the set of managed children, so place
 *      the children and change the grid widget size to reflect new size,
 *      if possible.
 *
 ************************************************************************/

static void
ChangeManaged( Widget w )
{
   XsmGridWidget gw = (XsmGridWidget) w;
   XsmGridWidgetClass gc = (XsmGridWidgetClass) XtClass(w);
   XtGeometryResult result = XtGeometryYes;
   XtWidgetGeometry geo_request, geo_reply;
   Boolean make_request = False;
   Dimension width, height;

  /* Update the col_width and row_height arrays and return the grid's
   * new height and width.
   */
   (*gc->grid_class.update_grid)(w, &width, &height);

  /* add in the margin widths and inter-column spacing */
   width += ((2 * gw->grid.margin_width) +
                      (gw->grid.h_spacing * (gw->grid.rows - 1)));

  /* add in the margin widths and inter-column spacing */
   height += ((2 * gw->grid.margin_height) +
                       (gw->grid.v_spacing * (gw->grid.columns - 1)));

  /* if the width changed request a new width */
   if (gw->core.width != width)
   {
      make_request = True;
      geo_request.request_mode |= CWWidth;
      geo_request.width = width;
   }

  /* if the height changed request a new height */
   if (gw->core.height != height)
   {
      make_request = True;
      geo_request.request_mode |= CWHeight;
      geo_request.height = height;
   }
   
   if (make_request)
   {
     /* Make a request from the grid's parent for a new size */
      result = XtMakeGeometryRequest(w, &geo_request, &geo_reply);

      if (result != XtGeometryYes)
      {
         int diff;

        /* Make a second request, which is guarenteed to succeeded
         * because the request is using the geo_reply values in the
         * request.
         */
         if (result == XtGeometryAlmost)
            result = XtMakeGeometryRequest(w, &geo_reply, NULL);

        /* if we couldn't change width */
         if (gw->core.width != width)
         {
            diff = (int) width - (int)gw->core.width;

           /* if couldn't grow, adjust margins and then clip if necessary */
            if (diff > 0)
            {
               /* first reduce margins as much as possible */
                if (gw->grid.margin_height >= diff)
                {
                   gw->grid.margin_height -= diff;
                }
                else
                {
                   diff -= gw->grid.margin_height;
                   gw->grid.margin_height = 0;
                }
            }
           /* if couldn't shrink, adjust margins */
            else
              gw->grid.margin_width = gw->grid.margin_width +
                                      (gw->core.width - width)/2;
         }

        /* if we couldn't grow in height */
         if (gw->core.height != height)
         {
            diff = (int) height - (int)gw->core.height;

           /* if couldn't grow, adjust margins and then clip if necessary */
            if (diff > 0)
            {
               /* first reduce margins as much as possible */
                if (gw->grid.margin_height >= diff)
                {
                   gw->grid.margin_height -= diff;
                }
                else
                {
                   diff -= gw->grid.margin_height;
                   gw->grid.margin_height = 0;
                }
            }
           /* if couldn't shrink, adjust margins */
            else
              gw->grid.margin_height += (gw->core.height - height)/2;
         }

        /* update the col_width and row_height to account for the adjustments */
         (*gc->grid_class.update_grid)(w, &width, &height);
      }
   }

  /* Layout children based on the adjustments made */
   (*gc->grid_class.layout_children)(w);

  /* Call motif routine that handles focus changes when children
   * are unmanaged.
   */
   _XmNavigChangeManaged(w);
}

/****************************************************************
 * Constraint Class Methods
 ****************************************************************/


/************************************************************************
 *
 *  ConstraintInitialize() -
 *      Like the Initialize procedure, this goes through the children's
 *      constraint resource values and determines their validity.  It
 *      also initializes necessary constraint fields in each child
 *        constraint record and updates the parents instance fields that
 *        are affected by changes in the children'c constraint resources. 
 *
 ************************************************************************/

static void
ConstraintInitialize( Widget req_w, Widget new_w, ArgList args,
                      Cardinal *num_args )
{
   XsmGridWidget parent = (XsmGridWidget) XtParent(new_w);
   XsmGridConstraintPart * grid_info = &(GridInfo(new_w)->grid);
   XsmGridWidgetClass gc = (XsmGridWidgetClass) XtClass((Widget)parent);
   Dimension row_pos = grid_info->row_pos;
   Dimension col_pos = grid_info->col_pos;
   Boolean auto_pos = False;

   grid_info->orig_row_pos = row_pos;
   grid_info->orig_col_pos = col_pos;

  /* Verify constraint resources */
   if (row_pos > parent->grid.rows - 1 && row_pos != MAXDIMENSION)
   {
      _XmWarning( (Widget) parent, MESSAGE3);
      row_pos = parent->grid.rows - 1;
   }

   if (col_pos > parent->grid.columns - 1 && col_pos != MAXDIMENSION)
   {
      _XmWarning( (Widget) parent, MESSAGE4);
      col_pos = parent->grid.columns - 1;
   }

   (*gc->grid_class.add_child)((Widget)parent, new_w, row_pos, col_pos);
}



/************************************************************************
 *
 *  ConstraintSetValues() -
 *      Like the SetValues procedure, this goes through the children's
 *      constraint resource values and determines their validity.
 *        It also updates the parents instance fields that are affected
 *        by changes in the children'c constraint resources. 
 *
 ************************************************************************/

static Boolean
ConstraintSetValues( Widget old_w, Widget request, Widget new_w, ArgList args,
                     Cardinal *num_args )
{
   XsmGridWidget parent = (XsmGridWidget) XtParent(new_w);
   XsmGridConstraintPart * new_grid_info = &(GridInfo(new_w)->grid);
   XsmGridConstraintPart * old_grid_info = &(GridInfo(new_w)->grid);
   Dimension col_pos = new_grid_info->col_pos;
   Dimension row_pos = new_grid_info->row_pos;
   Dimension old_col_pos = old_grid_info->col_pos;
   Dimension old_row_pos = old_grid_info->row_pos;
   Boolean redisplay = False;

   if (new_grid_info->row_pos > parent->grid.rows - 1)
   {
      _XmWarning( (Widget) parent, MESSAGE3);
      new_grid_info->row_pos = old_row_pos;
   }

   if (new_grid_info->col_pos > parent->grid.columns - 1)
   {
      _XmWarning( (Widget) parent, MESSAGE4);
      new_grid_info->col_pos = old_col_pos;
   }

   if (row_pos != old_row_pos || col_pos != old_col_pos)
   {
      if (parent->grid.widget_grid[row_pos][col_pos] == NULL)
      {
         parent->grid.widget_grid[row_pos][col_pos] = new_w;
         parent->grid.widget_grid[old_row_pos][old_col_pos] = NULL;
         redisplay = True;
      }
      else
      {
         _XmWarning( (Widget) parent, MESSAGE7);
         new_grid_info->row_pos = old_row_pos;
         new_grid_info->col_pos = old_col_pos;
      }
   }

   new_grid_info->orig_row_pos = new_grid_info->row_pos;
   new_grid_info->orig_col_pos = new_grid_info->col_pos;

   return (redisplay);
}


/****************************************************************
 * Grid Class Methods
 ****************************************************************/


/************************************************************
 *
 * UpdateGridDimensions() -
 *        Update the width and height of each row_height and
 *        col_width array and return the total width and height
 *        of the grid widget.
 *
 ************************************************************/

static void
UpdateGridDimensions(Widget w, Dimension * width, Dimension * height)
{
   XsmGridWidget gw = (XsmGridWidget) w;
   int i, j;

   *height = 0;
   *width = 0;

  /* update height for each row and keep track of total grid height */
   for (i = 0; i < gw->grid.rows; i++)
   {
      gw->grid.row_height[i] = GetRowHeight(gw, i);
      *height += gw->grid.row_height[i];
   }

  /* Initialize width for each column and keep track of total grid width */
   for (j = 0; j < gw->grid.columns; j++)
   {
      gw->grid.col_width[j] = GetColumnWidth(gw, j);
      *width += gw->grid.col_width[j];
   }
}



/******************************************************************
 *
 * AddChildToGrid() -
 *        Add the child to the grid based on the row and column
 *        positions.  If the position was not set track it as
 *        being quto set and initialize the position to 0.
 *
 ************************************************************/

static void
AddChildToGrid( Widget parent, Widget child,
                Dimension row_pos, Dimension col_pos)
{
   XsmGridWidget gw = (XsmGridWidget) parent;
   XsmGridConstraintPart * cur_grid_info;
   XsmGridConstraintPart * grid_info = &(GridInfo(child)->grid);
   Boolean place_column = False;
   Boolean place_row = False;
   Widget cur_child;

   if (row_pos == MAXDIMENSION && col_pos == MAXDIMENSION)
      grid_info->auto_placed = True;
   else
      grid_info->auto_placed = False; 

  /* if the row position was not originally set set them to 0 */
   if (row_pos == MAXDIMENSION) row_pos = 0;

  /* if the row position was not originally set set them to 0 */
   if (col_pos == MAXDIMENSION) col_pos = 0;

  /* Assign widgets to the widget_grid */
   if (gw->grid.widget_grid[row_pos][col_pos] == NULL)
   {
      gw->grid.widget_grid[row_pos][col_pos] = child;
   }
   else
   {
      /* If this widget has not been auto placed, replace any
       * widget that was auto placed and occupies this position.
       */
      if (!grid_info->auto_placed)
      {
	/* If the the row or column position were not set by
	 * by the grid widget, issue a warning message about the
	 * conflicting positions.
	 */
         if (grid_info->orig_col_pos != MAXDIMENSION &&
             grid_info->orig_row_pos != MAXDIMENSION)
            _XmWarning( parent, MESSAGE5);

         cur_child = gw->grid.widget_grid[row_pos][col_pos];
         cur_grid_info = &(GridInfo(cur_child)->grid);

        /* move the child that has been auto placed */
         if (cur_grid_info->auto_placed)
         {
            gw->grid.widget_grid[row_pos][col_pos] = cur_child;

            child = cur_child;
            grid_info = cur_grid_info;
         }
      }

      if (!grid_info->auto_placed) 
      {

        /* Try to place the child to a valid horizontal position
         * when the row_pos was set but the col_pos was not.
         */

         if (grid_info->orig_col_pos == MAXDIMENSION)
         {
            if (SetChildHorizPosition(gw, child, grid_info,
                                      row_pos, col_pos))
               return;
         }

        /* Try to place the child to a valid vertical position
         * when the col_pos was set but the row_pos was not.
         */

         if (grid_info->orig_row_pos == MAXDIMENSION)
         {
            if (SetChildVertPosition(gw, child, grid_info,
                                     row_pos, col_pos))
               return;
         }
      }

     /* Auto place the child to a valid position and issue
      * a warning message if the grid is full.
      */
      if (!SetChildPosition(gw, child, grid_info, row_pos, col_pos))
         _XmWarning( (Widget) parent, MESSAGE6);
   }
}


/******************************************************************
 *
 * LayoutChildren() -
 *        Run through the row and column arrays and move the
 *        children to their appropriate x,y position based on their
 *        location in the widget_grid array.  Align the children
 *        based on the setting of the child_alignment resource.
 *        This layout routine leaves holes for grid positions not
 *        filled with children, or filled with unmanaged children.
 *
 ******************************************************************/

static void
LayoutChildren(Widget w)
{
   XsmGridWidget gw = (XsmGridWidget) w;
   Widget child;
   Dimension col_adj, row_adj, col_width, row_height;
   Position x, y;
   int i, j;

  /* set the adjustment for the first row */
   row_adj = 0;

  /* Move through the rows, setting the position of the widget
   * in each column for each row.
   */
   for (i = 0; i < gw->grid.rows; i++)
   {
      /* set the adjustment for the first column */
       col_adj = 0;

      /* get the row height */
       row_height = gw->grid.row_height[i];

      /* set the position of the widget in each column in this row */
       for (j = 0; j < gw->grid.columns; j++)
       {
        /* get the child at this row/column location */
         child = gw->grid.widget_grid[i][j];

        /* get the column width */
         col_width = gw->grid.col_width[j];

        /* if there is a child at that location, move it to the appropiate
         * position.
         */
         if (child != NULL)
         {  
           /* adjust the x and y using margins, spacing, and adjustments */
            x = gw->grid.margin_width + (gw->grid.h_spacing * j) + col_adj;
            y = gw->grid.margin_height + (gw->grid.v_spacing * i) + row_adj;

           /* Adjust x,y if center_children resource is True and
            * if this is not the widest child in its current column.
            */
            if (child->core.width != col_width)
            {
               if (gw->grid.horiz_alignment == XsmALIGN_CENTER)
                  x += (col_width - child->core.width)/2; 
               else if (gw->grid.horiz_alignment == XsmALIGN_RIGHT)
                  x += (col_width - child->core.width);
            }

           /* Adjust x,y if center_children resource is True and
            * if this is not the tallest child in its current row.
            */
            if (child->core.height != row_height)
               if (gw->grid.vert_alignment == XsmALIGN_CENTER)
                  y += (row_height - child->core.height)/2; 
               else if (gw->grid.vert_alignment == XsmALIGN_BOTTOM)
                  y += (row_height - child->core.height);

           /* change the widgets location */
            if (child->core.x != x || child->core.y != y)
               _XmMoveObject (child, x, y);
         }

        /* Change adjustment for the next column */
         col_adj += col_width;
       }

      /* Change adjustment for the next row */
       row_adj += row_height;
   }
}

/******************************************************************
 *
 * DrawGridLines() -
 *     A grid class procedure that draws the lines in between the
 *     rows and columns of the grid.
 *
 ******************************************************************/

static void
DrawGridLines(Widget w)
{
   XsmGridWidget gw = (XsmGridWidget) w;

   ModifyGridLines(gw, True);
   gw->grid.lines_visible = True;
}

/******************************************************************
 *
 * EraseGridLines() -
 *     A grid class procedure that erases the lines in between the
 *     rows and columns of the grid.
 *
 ******************************************************************/

static void
EraseGridLines(Widget w)
{
   XsmGridWidget gw = (XsmGridWidget) w;

   ModifyGridLines(gw, False);
   gw->grid.lines_visible = False;
}


/****************************************************************
 * Internal Non-Class Functions
 ****************************************************************/
  
/************************************************************
 *
 * GetColumnWidth() -
 *        Get the width of the specified column by finding the
 *        width of the widest child.
 *
 ************************************************************/

static Dimension
GetColumnWidth(XsmGridWidget gw, int col_pos)
{
   Widget child;
   WidgetList * widget_grid = gw->grid.widget_grid;
   Dimension col_width = 0;
   int i;

   for (i = 0; i < gw->grid.rows; i++)

   {
       child = widget_grid[i][col_pos];

     /* Find the widest child in this column. */
       if (child != NULL)
          if (child->core.width > col_width)
             col_width = child->core.width;
   }

   return (col_width);
}


/************************************************************
 *
 * GetRowHeight() -
 *        Get the height of the specified row by finding the
 *        height of the tallest child.
 *
 ************************************************************/

static Dimension
GetRowHeight(XsmGridWidget gw, int row_pos)
{
   Widget child;
   WidgetList * widget_grid = gw->grid.widget_grid;
   Dimension row_height = 0;
   int i;

   for (i = 0; i < gw->grid.columns; i++)
   {
      child = widget_grid[row_pos][i];

     /* Find the tallest child in this row. */
      if (child != NULL)
         if (child->core.height > row_height)
            row_height = child->core.height;
   }

   return (row_height);
}



/******************************************************************
 *
 * SetChildHorizPosition() -
 *        Run through the current row to find a position that is
 *        not occupied, and assign that position to the child.
 *        Return False is a position if not found in the current row.
 *
 ************************************************************/

static Boolean
SetChildHorizPosition( XsmGridWidget parent, Widget child, 
                       XsmGridConstraintPart * grid_info,
                       Dimension row, Dimension start_col)
{
   XsmGridConstraintPart * cur_grid_info;
   Dimension cur_col;
   Widget cur_child;

   if (row >= parent->grid.rows)
      return False;

   if (start_col >= parent->grid.columns)
      start_col = parent->grid.columns - 1;

   cur_col = start_col++;

   while (cur_col != start_col)
   {
     /* If we are passed the last column in this row,
      * move to the first column.
      */
      if (cur_col == parent->grid.columns) cur_col = 0;

     /* if a position in the grid is not occupied by a widget
      * assign the child to this position and return True.  This
      * indicates that a position was found for the child.
      */
      if (parent->grid.widget_grid[row][cur_col] == NULL)
      {
         parent->grid.widget_grid[row][cur_col] = child;

        /* update grid_info */
         grid_info->row_pos = row;
         grid_info->col_pos = cur_col;

         return True;
      }

      cur_col++;
   }

  /* Return False is a place was not found in this row */
   return False;
}


/******************************************************************
 *
 * SetChildVertPosition() -
 *        Run through the current column to find a position that is
 *        not occupied, and assign that position to the child.
 *        Return False is a position if not found in the current
 *	  column.
 *
 ************************************************************/

static Boolean
SetChildVertPosition( XsmGridWidget parent, Widget child, 
                      XsmGridConstraintPart * grid_info,
                      Dimension start_row, Dimension col)
{
   XsmGridConstraintPart * cur_grid_info;
   Dimension cur_row;
   Widget cur_child;

   if (col >= parent->grid.columns)
      return False;

   if (start_row >= parent->grid.rows)
      start_row = parent->grid.rows - 1;

   cur_row = start_row++;

   while (cur_row != start_row)
   {
     /* If we are passed the last column in this row,
      * move to the first column.
      */
      if (cur_row == parent->grid.rows) cur_row = 0;

     /* if a position in the grid is not occupied by a widget
      * assign the child to this position and return True.  This
      * indicates that a position was found for the child.
      */
      if (parent->grid.widget_grid[cur_row][col] == NULL)
      {
         parent->grid.widget_grid[cur_row][col] = child;

        /* update grid_info */
         grid_info->row_pos = cur_row;
         grid_info->col_pos = col;

         return True;
      }

      cur_row++;
   }

  /* Return False is a place was not found in this row */
   return False;
}



/******************************************************************
 *
 * SetChildPosition() -
 *        Run through the widget grid to find a position that is
 *        not occupied, and assign that position to the child.  If
 *        a position is not found return False to indicate that the
 *        grid is full.
 *
 ************************************************************/

static Boolean
SetChildPosition( XsmGridWidget parent, Widget child, 
                  XsmGridConstraintPart * grid_info,
                  Dimension start_row, Dimension start_col)
{
   XsmGridConstraintPart * cur_grid_info;
   Dimension cur_col, cur_row;
   Widget cur_child;

   if (start_row >= parent->grid.rows)
      start_row = parent->grid.rows - 1;

   if (start_col >= parent->grid.columns)
      start_col = parent->grid.columns - 1;

   cur_row = start_row;
   cur_col = start_col;

   /* Search for next available grid position.  Begin at the column next
    * to the current position. Continue until we hit the end of the grid,
    * wrap to the beginning of the grid.  If we get back to the original
    * start position, break out of the loop.
    */
   while (cur_row < parent->grid.rows)
   {
      cur_col++;

     /* If we are passed the last column in this row,
      * move to the first column in the next row.
      */
      if (cur_col == parent->grid.columns)
      {
         cur_col = 0;
         cur_row++;
      }

     /* If we are passed the last row in this column,
      * move to the first row in the grid.
      */
      if (cur_row == parent->grid.rows)
         cur_row = 0;
      
     /* Break out of loop if the original start position is found. */
      if (cur_row == start_row && cur_col == start_col)
         break;

     /* if a position in the grid is not occupied by a widget
      * assign the child to this position and return True.  This
      * indicates that a position was found for the child.
      */
      if (parent->grid.widget_grid[cur_row][cur_col] == NULL)
      {
         parent->grid.widget_grid[cur_row][cur_col] = child;

        /* update grid_info */
         grid_info->row_pos = cur_row;
         grid_info->col_pos = cur_col;

         return True;
      }
      else
      {
         if (!grid_info->auto_placed)
         {
            cur_child = parent->grid.widget_grid[cur_row][cur_col];
            cur_grid_info = &(GridInfo(cur_child)->grid);

            if (cur_grid_info->auto_placed)
            {
               parent->grid.widget_grid[cur_row][cur_col] = cur_child;
               child = cur_child;
               grid_info = cur_grid_info;
            }
         }
      }
   }

  /* Since no position was found for the child, the grid must be full.
   * Return False to indicate a position was not found for the child.
   */

   return False;
}


static void
SetInvGC( XsmGridWidget gw, GC gc)
{
    unsigned long valueMask = (GCForeground | GCBackground);
    XGCValues values;

    values.foreground = gw->core.background_pixel;
    values.background = gw->manager.foreground;

    XChangeGC(XtDisplay((Widget)gw), gc, valueMask, &values);
}

static void
SetNormGC( XsmGridWidget gw, GC gc)
{
    unsigned long valueMask = (GCForeground | GCBackground);
    XGCValues values;

    values.background = gw->core.background_pixel;
    values.foreground = gw->manager.foreground;

    XChangeGC(XtDisplay((Widget)gw), gc, valueMask, &values);
}

static void
LoadGC( XsmGridWidget gw, Pixel fg, Pixel bg)
{
   unsigned long valueMask = (GCForeground | GCBackground);
   unsigned long dynamicMask;
   XGCValues values;

   if (gw->grid.gc != NULL)
       XtReleaseGC((Widget) gw, gw->grid.gc);

   values.background = gw->core.background_pixel;
   values.foreground = gw->manager.foreground;

   dynamicMask = (GCForeground | GCBackground);
   gw->grid.gc = XtAllocateGC((Widget) gw, gw->core.depth, valueMask,
                              &values, dynamicMask, 0);
}

static void
ModifyGridLines(XsmGridWidget gw,
#if NeedWidePrototypes
                int draw_grid)
#else
                Boolean draw_grid)
#endif /* NeedWidePrototypes */
{
   Widget widget = (Widget) gw;
   Dimension col_width, row_height;
   Position x, y, x2, y2;
   int i;

  /* Move through the columns, drawing a line between each column */

   /* Set y positions */
    y = gw->grid.margin_height - 1;
    y2 = (gw->core.height - gw->grid.margin_height) + 1;

   /* Initialize x position */
    x = gw->grid.margin_width;

    XDrawLine(XtDisplay(widget), XtWindow(widget), gw->grid.gc, x - 1, y,
							        x - 1, y2);

   /* set the position of the widget in each column in this row */
    for (i = 0; i < gw->grid.columns - 1; i++)
    {
     /* get the column width */
      col_width = gw->grid.col_width[i];

     /* add spacing, and col_width to x position */
      x += (gw->grid.h_spacing/2)+ 1 + col_width;

      if (draw_grid)
         SetNormGC(gw, gw->grid.gc);
      else
         SetInvGC(gw, gw->grid.gc);

      XDrawLine(XtDisplay(widget), XtWindow(widget), gw->grid.gc, x, y, x, y2);
    }

    x = (gw->core.width - gw->grid.margin_width) + 1;
    XDrawLine(XtDisplay(widget), XtWindow(widget), gw->grid.gc, x, y, x, y2);

  /* Move through the rows, drawing a line between each row */

   /* Set x positions */
   x = gw->grid.margin_width - 1;
   x2 = (gw->core.width - gw->grid.margin_width) + 1;

   /* Initialize y position */
   y = gw->grid.margin_height;

   XDrawLine(XtDisplay(widget), XtWindow(widget), gw->grid.gc, x, y - 1,
                                                              x2, y - 1);

   for (i = 0; i < gw->grid.rows - 1; i++)
   {
     /* get the row height */
      row_height = gw->grid.row_height[i];

     /* add spacing, and row_height to y position */
      y += (gw->grid.v_spacing/2) + 1 + row_height;

      if (draw_grid)
         SetNormGC(gw, gw->grid.gc);
      else
         SetInvGC(gw, gw->grid.gc);

      XDrawLine(XtDisplay(widget), XtWindow(widget), gw->grid.gc, x, y, x2, y);
   }

   y = (gw->core.height - gw->grid.margin_height) + 1;
   XDrawLine(XtDisplay(widget), XtWindow(widget), gw->grid.gc, x, y, x2, y);
}



/****************************************************************
 * Public Functions
 ****************************************************************/

/************************************************
 *
 * XsmCreateGrid -
 *     Grid widget creation convienence routine.
 *
 ************************************************/

Widget
XsmCreateGrid( Widget parent, char *name, ArgList arglist, Cardinal argcount )
{
    return (XtCreateWidget(name, xsmGridWidgetClass,
                           parent, arglist, argcount));
}



